import glob
import importlib
import inspect
import json
import sys

from pathlib import Path

def print_recipe_list():
    print("\nThe following built-in recipes are available:\n")

    # Find all metadata files.
    json_paths = glob.glob(str(Path(__file__).parent / '**' / 'metadata.json'))
    json_paths.sort()

    for json_path in json_paths:
        with open(json_path) as f:
            metadata = json.load(f)
            if 'script_name' not in metadata:
                continue
            script_name = metadata['script_name']
            display_name = metadata.get('display_name', 'NO DISPLAY NAME')
            print(f"  {script_name} -- {display_name}")

def get_recipe_cls(script_name, pkg_recipe):
    # Get all classes defined in the recipe package.
    for _, value in inspect.getmembers(pkg_recipe, inspect.isclass):
        # Check if the current class is a subclass of Recipe and has the
        # requested script name.
        if (issubclass(value, pkg_recipe.Recipe) and
            value is not pkg_recipe.Recipe and
            value.get_metadata('script_name') == script_name):
            return value

    sys.exit("ERROR: Unknown recipe.")

def get_parsed_args(recipe_cls):
    parser = recipe_cls.get_argument_parser()
    parser.add_context_arguments()
    return parser.parse_args(sys.argv[2:])

def run_recipe(script_name, pkg_recipe):
    recipe_cls = get_recipe_cls(script_name, pkg_recipe)
    parsed_args = get_parsed_args(recipe_cls)

    try:
        with pkg_recipe.Context.create_context(parsed_args.mode) as context:
            with recipe_cls(parsed_args) as recipe:
                recipe.run(context)
                if recipe.output_dir is not None:
                    print(f"Generated:\n    {recipe.get_output_dir()}")
                else:
                    print("SKIPPED: Nothing to do.")
    except pkg_recipe.ModeModuleNotFoundError as e:
        req_file = pkg_recipe.nsys_constants.NSYS_RECIPE_REQ_DASK_PATH
        sys.exit(f"ERROR: {e}\nPlease install packages from '{req_file}'")
    except pkg_recipe.StatsModuleNotFoundError as e:
        sys.exit(f"ERROR: {e}\nPlease make sure the Stats report exists.")
    except pkg_recipe.StatsInternalError as e:
        sys.exit(f"ERROR: {e}\nStats encountered an internal error.")
    except pkg_recipe.ValueError as e:
        sys.exit(f"ERROR: {e}")
    except pkg_recipe.NoDataError:
        # Early exit due to lack of data.
        sys.exit("SKIPPED: No data available.")

def main():
    if len(sys.argv) < 2:
        sys.exit("ERROR: No recipe specified.")

    if '--help-recipes' in sys.argv:
        print_recipe_list()
        sys.exit(0)

    try:
        pkg_recipe = importlib.import_module('nsys_recipe')
    except ModuleNotFoundError as e:
        sys.exit(f"ERROR: {e}\nPlease add the nsys_recipe directory to PYTHONPATH.")
    else:
        run_recipe(sys.argv[1], pkg_recipe)

if __name__ == "__main__":
    main()
