import importlib
import numpy as np
import os
import pandas as pd

from collections import OrderedDict
from datetime import datetime

from nsys_recipe.diff_sum import diff_sum
from nsys_recipe.lib import helpers
from nsys_recipe.lib import recipe
from nsys_recipe.lib.args import Option

class CudaApiSum(recipe.Recipe):
    @staticmethod
    def mapper_func(nsysrep, parsed_args):
        sqlite_file = helpers.nsysrep_to_sqlite(nsysrep)
        if sqlite_file is None:
            return None

        stats_cls = helpers.get_stats_cls('cuda_api_sum', 'CudaApiSum')
        return helpers.stats_cls_to_df(sqlite_file, parsed_args, stats_cls)

    def aggregate_stats(self, df):
        grouped = df.groupby(df.index.name)

        d = OrderedDict()
        total_time = grouped['Total Time'].sum()
        d['Time'] = (total_time / total_time.sum() * 100)
        d['Total Time'] = total_time
        d['Num Calls'] = grouped['Num Calls'].sum()
        d['Avg'] = d['Total Time'] / d['Num Calls']
        d['Q1'] = grouped['Q1'].min()
        d['Med'] = grouped['Med'].median()
        d['Q3'] = grouped['Q3'].max()
        d['Min'] = grouped['Min'].min()
        d['Max'] = grouped['Max'].max()
        d['StdDev'] = grouped.apply(
            lambda x: helpers.stddev(x, d, 'Num Calls'))
        min_value = grouped['Min'].min()
        d['Min Report'] = grouped.apply(
            lambda x: ', '.join(x.loc[x['Min'] == min_value.loc[x.name], 'Report']))
        max_value = grouped['Max'].max()
        d['Max Report'] = grouped.apply(
            lambda x: ', '.join(x.loc[x['Max'] == max_value.loc[x.name], 'Report']))

        aggregated_df = pd.concat(d.values(), axis=1, keys=d.keys()).round(1)
        return aggregated_df.sort_values(by=['Total Time'], ascending=False)

    def reducer_func(self, dfs):
        dfs = helpers.filter_none(dfs)
        df = pd.concat(dfs)

        # Remove any tags or hidden columns that are for internal use.
        df.columns = df.columns.str.replace('(:).*', '', regex=True)
        df.columns = df.columns.str.lstrip('_')

        df = df.set_index('Name')
        df = df[['Time', 'Total Time', 'Num Calls', 'Avg', 'Q1', 'Med', 'Q3',
            'Min', 'Max', 'StdDev', 'Report']]

        per_rank_df = df.sort_values(by=['Report', 'Total Time'], ascending=[True, False])
        all_ranks_df = self.aggregate_stats(df)

        per_rank_df.to_parquet(self.add_output_file('per-rank.parquet'))
        all_ranks_df.to_parquet(self.add_output_file('all-ranks.parquet'))

        if self._parsed_args.csv:
            per_rank_df.to_csv(self.add_output_file('per-rank.csv'))
            all_ranks_df.to_csv(self.add_output_file('all-ranks.csv'))

    def save_notebook(self):
        self.create_notebook('stats.ipynb')
        self.add_notebook_helper_file('nsys_display.py')

    def save_analysis_file(self):
        self._analysis_dict.update({
            'EndTime': str(datetime.now()),
            'InputReports': self._parsed_args.input,
            'Outputs': self._output_files
        })
        self.create_analysis_file()

    def run(self, context):
        super().run(context)
        if self._parsed_args.diff:
            diff_sum.run(self, context)
            return

        mapper_res = context.wait(context.map(
            self.mapper_func,
            self._parsed_args.input,
            parsed_args=self._parsed_args
        ))
        self.reducer_func(mapper_res)

        self.save_notebook()
        self.save_analysis_file()

    @classmethod
    def get_argument_parser(cls):
        parser = super().get_argument_parser()

        mutually_exclusive_group = parser.recipe_group.add_mutually_exclusive_group(required=True)
        parser.add_argument_to_group(mutually_exclusive_group, Option.REPORT_DIR)
        parser.add_argument_to_group(mutually_exclusive_group, Option.DIFF)
        parser.add_argument_to_group(mutually_exclusive_group, Option.INPUT)

        parser.add_recipe_argument(Option.OUTPUT)
        parser.add_recipe_argument(Option.FORCE_OVERWRITE)
        parser.add_recipe_argument(Option.AUTO_INCREMENT)
        parser.add_recipe_argument(Option.START)
        parser.add_recipe_argument(Option.END)
        parser.add_recipe_argument(Option.CSV)

        return parser
