import pandas as pd

import nsysstats

from nsys_recipe.lib import exceptions

class TableConfig:
    """Configuration used for reading tables from a database.

    Attributes:
    - columns (list): name of columns to read.
    - optional (bool): whether the table is optional.
    - time_filter (bool): whether the table needs to be time filtered.
    - pre_filter_func (callable): callable filtering function that takes the
        dataframe as input (applied before the time filtering).
    """
    def __init__(self, columns=None, optional=False, time_filter=False, pre_filter_func=None):
        self.columns = columns or []
        self.optional = optional
        self.time_filter = time_filter
        self.pre_filter_func = pre_filter_func

def _read_table_to_dataframe(conn, table, columns):
    column_query = ','.join(columns) if columns else '*'
    query = f"SELECT {column_query} FROM {table}"
    return pd.read_sql(query, conn)

def _get_time_cols(df):
    if 'start' in df.columns:
        if 'end' in df.columns:
            # Time range.
            return ('start', 'end')
        else:
            # Point in time.
            return ('start', 'start')

    if 'timestamp' in df.columns:
        # Point in time.
        return ('timestamp', 'timestamp')

    # Update the function in case a new time column is added.
    raise NotImplementedError

def _filter_by_time_range(df, start_time, end_time):
    start_col, end_col = _get_time_cols(df)

    mask = pd.Series(True, index=df.index)
    if end_time is not None:
        mask &= (df[start_col] <= end_time)
    if start_time is not None:
        mask &= (df[end_col] >= start_time)

    return df[mask]

def read_sqlite_tables(db, table_config_dict, start_time, end_time):
    """Read tables from an sqlite database.

    Note the following details for each table's 'TableConfig':
    - Only the columns specified in the 'columns' list will be read. If the
      list is empty, all columns will be read.
    - If 'optional' is set to False, an error message will be printed and the
      function will return None if the table or a column is not found.
    - If 'time_filter' is set to True, the dataframe will be filtered,
      retaining only events that start or end within the given range.
    - If 'pre_filter_func' is set, it will be applied to the dataframe
      before the time filtering.

    Parameters
    ----------
    db : str
        Database path.
    table_config_dict : dict
        Dictionary containing the TableConfig for each table.
    start_time : int
        The start time of the desired range.
    end_time : int
        The end time of the desired range.

    Returns
    -------
    df_dict : dict
        Dictionary containing the table name and the corresponding dataframe.
    """
    # Use nsysstats Report class to manage the sqlite3 connection.
    try:
        sql_handler = nsysstats.Report(db)
    except Exception as e:
        raise exceptions.StatsInternalError(e) from e

    df_dict = {}
    for table, config in table_config_dict.items():
        if not sql_handler.table_exists(table):
            if not config.optional:
                print(f"{db} does not contain the '{table}' table.")
                return None
            continue

        for column in config.columns:
            if not sql_handler.table_col_exists(table, column):
                if not config.optional:
                    print(f"{db} does not contain the '{column}' column."
                        " Please re-export the report file with a recent version of Nsight Systems.")
                    return None
                continue

        df = _read_table_to_dataframe(sql_handler.dbcon, table, config.columns)

        if config.pre_filter_func:
            df = config.pre_filter_func(df)
        if config.time_filter:
            df = _filter_by_time_range(df, start_time, end_time)

        df_dict[table] = df

    return df_dict

def replace_id_with_value(main_df, str_df, id_column, new_column_name=None):
    """Replace the values in 'id_column' of 'main_df' with the corresponding
    string value stored in 'str_df'.

    Parameters
    ----------
    main_df : dataframe
        Dataframe containing 'id_column'.
    str_df : dataframe
        Dataframe 'StringId' that maps IDs to string values.
    id_column : str
        Name of the column that should be replaced with the corresponding
        string values.
    new_column_name :  str
        New name for the replaced column. If not provided, the original column
        name will be used.
    """
    renamed_str_df = str_df.rename(columns={'id': id_column})
    merged_df = main_df.merge(renamed_str_df, on=id_column, how='left')

    # Drop the original 'id_column' column and rename 'value' column to 'id_column'.
    merged_df = merged_df.drop(columns=[id_column])

    if new_column_name is None:
        new_column_name = id_column
    merged_df = merged_df.rename(columns={'value': new_column_name})

    return merged_df

def extract_pid(global_id_series):
    """Extract the PID from the global ID.

    Parameters
    ----------
    global_id_series: series
        Either the 'globalPid' or 'globalTid' column of the dataframe.
    """
    return global_id_series // 0x1000000 % 0x1000000
