import contextlib
import pandas as pd

from datetime import datetime
from functools import partial
from pathlib import Path

from nsys_recipe.lib import helpers
from nsys_recipe.lib import loader
from nsys_recipe.lib import nvtx
from nsys_recipe.lib import recipe
from nsys_recipe.lib import summary

from nsys_recipe.lib.args import Option
from nsys_recipe.lib.loader import TableConfig

class MpiSum(recipe.Recipe):
    @staticmethod
    def mapper_func(nsysrep, parsed_args):
        sqlite_file = helpers.nsysrep_to_sqlite(nsysrep)
        if sqlite_file is None:
            return None

        table_config_dict = {
            'StringIds': TableConfig(),
            'NVTX_EVENTS': TableConfig(
                ['text', 'start', 'end', 'textId', 'domainId', 'eventType'],
                optional=True,
                time_filter=True,
                pre_filter_func=partial(nvtx.get_domain_range_df, domain_name='MPI')
            ),
            'MPI_P2P_EVENTS': TableConfig(
                ['start', 'end', 'textId'],
                optional=True,
                time_filter=True
            ),
            'MPI_START_WAIT_EVENTS': TableConfig(
                ['start', 'end', 'textId'],
                optional=True,
                time_filter=True
            ),
            'MPI_OTHER_EVENTS': TableConfig(
                ['start', 'end', 'textId'],
                optional=True,
                time_filter=True
            ),
            'MPI_COLLECTIVES_EVENTS': TableConfig(
                ['start', 'end', 'textId'],
                optional=True,
                time_filter=True
            )
        }

        df_dict = loader.read_sqlite_tables(
            sqlite_file,
            table_config_dict,
            parsed_args.start,
            parsed_args.end
        )
        if df_dict is None:
            return None

        range_dfs = []

        if 'NVTX_EVENTS' in df_dict:
            range_df = nvtx.combine_text_fields(df_dict['NVTX_EVENTS'], df_dict['StringIds'])
            if not range_df.empty:
                range_dfs.append(range_df)

        with contextlib.suppress(Exception):
            mpi_dfs = pd.concat([
                df_dict.get('MPI_P2P_EVENTS'),
                df_dict.get('MPI_START_WAIT_EVENTS'),
                df_dict.get('MPI_OTHER_EVENTS'),
                df_dict.get('MPI_COLLECTIVES_EVENTS')
            ])
            range_df = loader.replace_id_with_value(
                mpi_dfs, df_dict['StringIds'], 'textId', 'text')
            range_dfs.append(range_df)

        if not range_dfs:
            print(f"{sqlite_file} does not contain MPI data.")
            return None

        range_df = pd.concat(range_dfs, join='inner')

        range_df['duration'] = range_df['end'] - range_df['start']
        range_df = range_df[range_df['duration'] > 0]
        range_gdf = range_df[['text', 'duration']].groupby('text')

        filename = Path(sqlite_file).stem
        stats_df = summary.describe_duration(range_gdf['duration'])

        return summary.StatsInfo(filename, stats_df)

    def reducer_func(self, mapper_res):
        stats_infos = helpers.filter_none(mapper_res)
        stats_infos = sorted(stats_infos, key=lambda x: x.filename)

        filenames, stats_dfs = zip(*stats_infos)

        rank_file_df = pd.DataFrame(filenames, columns=['File'])
        rank_file_df.to_parquet(self.add_output_file('files.parquet'), index=False)

        all_stats_df = summary.aggregate_stats_dfs(stats_dfs)
        all_stats_df.to_parquet(self.add_output_file('all_stats.parquet'))

        rank_stats_dfs = [df.assign(Rank=rank, File=filenames[rank]) for rank, df in enumerate(stats_dfs)]
        rank_stats = pd.concat(rank_stats_dfs)
        rank_stats.to_parquet(self.add_output_file('rank_stats.parquet'))

        if self._parsed_args.csv:
            all_stats_df.to_csv(self.add_output_file('all_stats.csv'))
            rank_stats.to_csv(self.add_output_file('rank_stats.csv'))

    def save_notebook(self):
        self.create_notebook('stats.ipynb')
        self.add_notebook_helper_file('nsys_pres.py')

    def save_analysis_file(self):
        self._analysis_dict.update({
            'EndTime': str(datetime.now()),
            'InputReports': self._parsed_args.input,
            'Outputs': self._output_files
        })
        self.create_analysis_file()

    def run(self, context):
        super().run(context)
        mapper_res = context.wait(context.map(
            self.mapper_func,
            self._parsed_args.input,
            parsed_args=self._parsed_args
        ))
        self.reducer_func(mapper_res)

        self.save_notebook()
        self.save_analysis_file()

    @classmethod
    def get_argument_parser(cls):
        parser = super().get_argument_parser()

        mutually_exclusive_group = parser.recipe_group.add_mutually_exclusive_group(required=True)
        parser.add_argument_to_group(mutually_exclusive_group, Option.REPORT_DIR)
        parser.add_argument_to_group(mutually_exclusive_group, Option.INPUT)

        parser.add_recipe_argument(Option.OUTPUT)
        parser.add_recipe_argument(Option.FORCE_OVERWRITE)
        parser.add_recipe_argument(Option.AUTO_INCREMENT)
        parser.add_recipe_argument(Option.START)
        parser.add_recipe_argument(Option.END)
        parser.add_recipe_argument(Option.CSV)

        return parser
